

################################################################################
## This function prepares an exprSet class object.
## input:
## arr1: Filtered logged and normalized gene expression matrix for current analysis
## pheno: response variable
## experiment: array or sample or expriment id
## filepath: path for storing binary file
## output:
## create and save an exprSet object: aExprSet to disk
################################################################################
#' @import Biobase
CreateExprSetData <- function(arr1,
                              pheno,
                              experiment,
                              filepath) {

#  source(paste(ArrayToolsPath,"/R/BiocUtil.R", sep=""), echo = FALSE)
  #print("Loading R package: Biobase ...")
  #if (!LoadRPackage("Biobase", isBioconductor=T)) return(FALSE)
  ip <- installed.packages()[, "Package"]
  if (! "Biobase" %in% ip) return(FALSE)

  #validity check
  if( ncol(arr1) != nrow(pheno) ) {
    print("ERROR: the number of columns in expression matrix is not equal to the number of rows in covariates dataframe")
    return(FALSE)
  }

  print("Creating ExpressionSet object ...")
  #note:
    #1.use the following names for survival phenotype: TIMEsurvival, EVENTdeath.
    #EVENTdeath=1 indicates occurrence of event, 0 otherwise. Will use in ConductGlobalTest() function
    #2.categorical variables should be explicitly coded as factor
  pheno <- data.frame(pheno)
  experiment <- data.frame(experiment)
  if (length(pheno)==1) {
    colnames(pheno) <- "CLASS"
    rownames(pheno) <- experiment[[1]]
    if (!(class(pheno$CLASS)=="factor")) pheno$CLASS <- as.factor(pheno$CLASS)
    metadata <- data.frame(labelDescription=c("class variable"), row.names=c("CLASS"))
  } else {
    colnames(pheno) <- c("EVENTdeath", "TIMEsurvival")
    rownames(pheno) <- experiment[[1]]
    metadata <- data.frame(labelDescription=c("survival status, 0 = No, 1 = Yes", "survival time"), row.names=c("EVENTdeath", "TIMEsurvival"))
  }

  #generate expression level matrix
    #1. arr1 is FILTERED logged normalized gene expression matrix
    #2. the indices of the FILTERED genes is used as the rownames for aExprMat matrix. This feature is later used in conducting GlobalTest.
  aExprMat <- as.matrix(arr1)
  colnames(aExprMat) <- experiment[[1]]
  rownames(aExprMat) <- 1:nrow(aExprMat)

  #generate phenoData class object
  aPheno <- new("AnnotatedDataFrame", data=pheno, varMetadata=metadata)

  #generate exprSet class object
  aExprSet <- new("ExpressionSet", exprs=aExprMat, phenoData=aPheno)

  print("Saving ExpressionSet object to disk")
  #save aExprSet to disk
  save(aExprSet, file=paste(filepath, "/ExprSet.rda", sep=""))

  return(TRUE)
}
