
#' Gene set enrichment analysis
#'
#' This function conducts gene set enrichment analysis among pre-defined classes and for survival data and quantitative
#' trait data, respectively. It finds BioCarta pathways, KEGG pathways, transcription factor target lists
#' or microRNA target lists with statistically significant differences among pre-defined classes.
#' It aslo finds gene sets that are correlated with survival or quantitative trait of samples.
#' A gene set is selected if its corresponding re-sampling p-value is below the specified threshold. The re-sampling p-value
#' is calculated through permutation tests.
#' Basically, 100,000 LS (log score) or KS (Kolmogorov-Smirnov) permutation tests are
#' conducted to calculate a p-value measuring
#' the gene set enrichment. For each gene set, N genes are randomly selected from a
#' gene list in analysis,
#' where N is the number of genes belonging to that gene set
#' For each permutation, the LS or KS statistics associated with that gene set are
#' computed based on the p-value.
#' The p-value for that gene set is then defined as the proportion of permutations
#' for which the LS (KS) statistics are larger than the observed LS (KS) statistics
#' from original data.
#' @param expr matrix of gene expression data for training samples. Rows are genes and columns are arrays.
#' Its column names must be provided.
#' @param filter vector of 1's or 0's of the same length as genes. 1 means to keep the gene
#' while 0 means to exclude genes
#' from the gene set enrichment analysis. If \code{rvm = TRUE}, all genes will be used
#' in random variance model estimation.
#' @param surv logical specifying if it is survival data or not. Default is \code{FALSE}.
#' @param time vector specifying survival time. Defualt is  \code{NULL}.
#' @param status vector specifying survival status. Defualt is \code{NULL}.
#' @param quant vector specifying the quantitative trait of samples. Default is \code{NULL}.
#' @param geneId matrix/data frame of gene IDs. Rows are IDs and columns are annotations
#' such as Symbol and EntrezID. Its row names must be provided as IDs, and one of its
#' column names must represent gene symbols. The rows must be in the same order as
#' those of IDs in \code{expr}.
#' @param cls vector of sample classes.
#' @param isPaired logical. If \code{rvm = TRUE}, samples are paired.
#' @param pairID vector of pairing variables for all samples.
#' @param doGroupComparison logical. If \code{TRUE}, it compares enrichment between two groups samples.
#' Default is \code{FALSE}.
#' @param grpID vector specifying the group ID. Default is \code{NULL}.
#' @param hasDuplic logical. If \code{rvm = TRUE}, array replicates will be averaged. Default is FALSE.
#' @param duplicID vector specifying array replicates. Default is \code{NULL}.
#' @param rvm logical. If \code{TRUE}, random variance model will be employed.
#' Default is \code{TRUE}.
#' @param nperm.GSA numeric specifying the number of permutation tests in \code{GSA()} function. Defualt is 200.
#' @param geneSetType vector specifying the type of gene sets for enrichment analysis. It can be "BioCarta" , "KEGG", "TF"
#'  and "microRNA", representing BioCarta pathways, KEGG pathways, experimentally verified transcription factor
#'  targets and experimentally verified microRNA targets, respectively.
#' @param fromKEGGdb logical specifying if KEGG pathways are obtained from the KEGG.db or KEGGREST R package.
#'  When \code{fromKEGGdb = TRUE}, the 229 KEGG pathways from the KEGG.db R package are used.
#'  When \code{fromKEGGdb = FALSE}, the 333 KEGG pathways obtained from the KEGGREST R
#' package are used. Default is \code{TRUE}.
#' @param frommiRTarBasev4 logical specifying if the microRNA target lists are obtained from the miRTarBase database v4 or v7
#' at http://mirtarbase.mbc.nctu.edu.tw/php/download.php. When \code{frommiRTarBasev4 = TRUE},
#' the microRNA target lists from miRTarBase v4.x are used. When \code{frommiRTarBasev4 = FALSE},
#' the microRNA target lists from miRTarBase v7.0 are used. Default is \code{TRUE}.
#' @param pathwayMin the minimal number of genes being allowed in one pathway.
#' Default is 5.
#' @param pathwayMax the maximal number of genes being allowed in one pathway.
#' Default is 2000.
#' @param isSingleChannel logical. If \code{TRUE}, data are single-channel; otherwsie, they are two-channel. Default
#' is \code{TRUE}.
#' @param alpha numeric specifying the significant level being set for significantly enriched pathways.
#' Default is 0.05.
#' @param seed numeric specifying random seed for permutation tests. Default is 123456.
#' @param organism character specifying the organism. It can be "human" or "mouse". When \code{doBiocarta = TRUE},
#' only "human" is available for \code{organism}.
#' @param corrtest.method charcater specifying the Pearson or Spearman correlation test to find enriched gene sets
#' given the quantitative trait data. It can be "pearson" or "spearman".
#' @param projectPath character specifying the ouput directory.
#' @param outputName character specifying the directory for keeping the HTML document and associated results.
#' Default is "GeneSetClassComparison".
#' @param popHTML logical. If TRUE, an HTML document with significantly enriched pathways will be popped up. The file will
#' be saved as <projectPath>/Output/<outputName>/<outputName>.html.
#' Default is \code{TRUE}.

#' @return a data frame containing the enriched BioCarta pathways, KEGG pathways, transcription factor target lists
#' or microRNA target lists, pathway description, number
#' of genes in each enriched pathway, p-values for LS/KS permutation tests and Efron-Tibshirani's GSA tests.
#' @author BRB-ArrayTools Development Team, \email{arraytools@@emmes.com}
#' @references Xu X, Zhao Y and Simon R. Gene Set Expression Comparison kit for BRB-ArrayTools. Bioinformatics 2008. 24: 137-9.
#'  \cr BRB-ArrayTools manual: \url{https://brb.nci.nih.gov/BRB-ArrayTools/Documentation.html}.

#' @export
#' @useDynLib GSEA geneSetEnrichmentQT
#' @examples
#' ## find BioCarta significant pathways among two classes in a breast cancer dataset
#' dataset<-"Brca"
#' # gene IDs
#' geneId <- read.delim(system.file("extdata", paste0(dataset, "_GENEID.txt"),
#'                                  package = "GSEA"), as.is = TRUE,
#'                                  colClasses = "character")
#' # gene expression
#' x <- read.delim(system.file("extdata", paste0(dataset, "_LOGRAT.TXT"),
#'                             package = "GSEA"), header = FALSE)
# filter information: 1 - pass the filter, 0 - filtered
#' filter <- scan(system.file("extdata", paste0(dataset, "_FILTER.TXT"),
#'                            package = "GSEA"), quiet = TRUE)
#' # sample information
#' expdesign <- read.delim(system.file("extdata", paste0(dataset, "_EXPDESIGN.txt"),
#'                                     package = "GSEA"), as.is = TRUE)
#' ind1 <- which(expdesign[, 4] == "BRCA1")
#' ind2 <- which(expdesign[, 4] == "BRCA2")
#' ind <- c(ind1, ind2)
#' expr <- x[, ind]
#' colnames(expr) <- expdesign[ind, 1]
#' projectPath <- file.path(Sys.getenv("HOME"),"Brca")
#' outputName <- "GeneSetClassComparison"
#' cls <- c(rep("BRCA1", length(ind1)), rep("BRCA2", length(ind2)))
#' gsea(expr = expr,
#'      filter = filter,
#'      geneId = geneId,
#'      cls = cls,
#'      geneSetType = "BioCarta",
#'      isSingleChannel = FALSE,
#'      alpha = 0.005,
#'      organism = "human",
#'      projectPath = projectPath,
#'      outputName = outputName)
#'
#' ## find BioCarta pathways that are significantly correlated with survival
#' dataset<-"Pomeroy"
#' # gene IDs
#' geneId <- read.delim(system.file("extdata", paste0(dataset, "_GENEID.txt"),
#'                                  package = "GSEA"), as.is = TRUE,
#'                                  colClasses = "character")
#' # expression data
#' x <- read.delim(system.file("extdata", paste0(dataset, "_LOGINT.TXT"),
#'                             package = "GSEA"), header = FALSE)
#' # filter information, 1 - pass the filter, 0 - filtered
#' filter <- scan(system.file("extdata", paste0(dataset, "_FILTER.TXT"),
#'                            package = "GSEA"), quiet = TRUE)
#' # sample information
#' expdesign <- read.delim(system.file("extdata", paste0(dataset, "_EXPDESIGN.txt"),
#'                                    package = "GSEA"), as.is = TRUE)
#' time <- expdesign[,7]
#' status <- expdesign[,12]
#' ind1 <- which(status == 0)
#' ind2 <- which(status == 1)
#' ind <- c(ind1, ind2)
#' expr <- x[, ind]
#' time <- time[ind]
#' status <- status[ind]
#' colnames(expr) <- expdesign[ind, 1]
#' projectPath <- file.path(Sys.getenv("HOME"),dataset)
#' outputName <- "GeneSetSurvivalComparison"
#' gsea(expr = expr,
#'      filter = filter,
#'      surv = T,
#'      time = time,
#'      status = status,
#'      geneId = geneId,
#'      rvm = FALSE,
#'      geneSetType = "BioCarta",
#'      alpha = 0.005,
#'      organism = "human",
#'      projectPath = projectPath,
#'      outputName = outputName)
#'
#' ## find KEGG pathways that are significally correlated with simulated
#' ## quantitative trait of samples
#' dataset<-"Brca"
#' # gene IDs
#' geneId <- read.delim(system.file("extdata", paste0(dataset, "_GENEID.txt"),
#'                                  package = "GSEA"), as.is = TRUE, colClasses = "character")
#' # expression data
#' expr <- read.delim(system.file("extdata", paste0(dataset, "_LOGRAT.TXT"),
#'                                package = "GSEA"), header = FALSE)
#' # filter information, 1 - pass the filter, 0 - filtered
#' filter <- scan(system.file("extdata", paste0(dataset, "_FILTER.TXT"),
#'                            package = "GSEA"), quiet = TRUE)
#' # sample information
#' expdesign <- read.delim(system.file("extdata", paste0(dataset, "_EXPDESIGN.txt"),
#'                                     package = "GSEA"), as.is = TRUE)
#' quant <- expdesign[,11]
#' colnames(expr) <- expdesign[, 1]
#' projectPath <- file.path(Sys.getenv("HOME"),"Brca")
#' outputName <- "GeneSetQTComparison"
#' gsea(expr = expr,
#'      filter = filter,
#'      quant = quant,
#'      geneId = geneId,
#'      geneSetType = "KEGG",
#'      isSingleChannel = FALSE,
#'      alpha = 0.005,
#'      organism = "human",
#'      corrtest.method = "pearson",
#'      projectPath = projectPath,
#'      outputName = outputName)
#'
#' ## find significant TF target gene lists among two classes in a breast cancer dataset
#' dataset<-"Brca"
#' # gene IDs
#' geneId <- read.delim(system.file("extdata", paste0(dataset, "_GENEID.txt"),
#'                                  package = "GSEA"), as.is = TRUE,
#'                                  colClasses = "character")
#' # gene expression
#' x <- read.delim(system.file("extdata", paste0(dataset, "_LOGRAT.TXT"),
#'                             package = "GSEA"), header = FALSE)
# filter information: 1 - pass the filter, 0 - filtered
#' filter <- scan(system.file("extdata", paste0(dataset, "_FILTER.TXT"),
#'                            package = "GSEA"), quiet = TRUE)
#' # sample information
#' expdesign <- read.delim(system.file("extdata", paste0(dataset, "_EXPDESIGN.txt"),
#'                                     package = "GSEA"), as.is = TRUE)
#' ind1 <- which(expdesign[, 4] == "BRCA1")
#' ind2 <- which(expdesign[, 4] == "BRCA2")
#' ind <- c(ind1, ind2)
#' expr <- x[, ind]
#' colnames(expr) <- expdesign[ind, 1]
#' projectPath <- file.path(Sys.getenv("HOME"),"Brca")
#' outputName <- "GeneSetClassComparison"
#' cls <- c(rep("BRCA1", length(ind1)), rep("BRCA2", length(ind2)))
#' gsea(expr = expr,
#'      filter = filter,
#'      geneId = geneId,
#'      cls = cls,
#'      geneSetType = "TF",
#'      isSingleChannel = FALSE,
#'      alpha = 0.005,
#'      organism = "human",
#'      projectPath = projectPath,
#'      outputName = outputName)
#'
#'## find significant microRNA target gene lists among two classes
#'## in a breast cancer dataset
#' dataset<-"Brca"
#' # gene IDs
#' geneId <- read.delim(system.file("extdata", paste0(dataset, "_GENEID.txt"),
#'                                  package = "GSEA"), as.is = TRUE,
#'                                  colClasses = "character")
#' # gene expression
#' x <- read.delim(system.file("extdata", paste0(dataset, "_LOGRAT.TXT"),
#'                             package = "GSEA"), header = FALSE)
# filter information: 1 - pass the filter, 0 - filtered
#' filter <- scan(system.file("extdata", paste0(dataset, "_FILTER.TXT"),
#'                            package = "GSEA"), quiet = TRUE)
#' # sample information
#' expdesign <- read.delim(system.file("extdata", paste0(dataset, "_EXPDESIGN.txt"),
#'                                     package = "GSEA"), as.is = TRUE)
#' ind1 <- which(expdesign[, 4] == "BRCA1")
#' ind2 <- which(expdesign[, 4] == "BRCA2")
#' ind <- c(ind1, ind2)
#' expr <- x[, ind]
#' colnames(expr) <- expdesign[ind, 1]
#' projectPath <- file.path(Sys.getenv("HOME"),"Brca")
#' outputName <- "GeneSetClassComparison"
#' cls <- c(rep("BRCA1", length(ind1)), rep("BRCA2", length(ind2)))
#' gsea(expr = expr,
#'      filter = filter,
#'      geneId = geneId,
#'      cls = cls,
#'      geneSetType = "microRNA",
#'      isSingleChannel = FALSE,
#'      alpha = 0.005,
#'      organism = "human",
#'      projectPath = projectPath,
#'      outputName = outputName)

gsea <- function(expr,
                 filter,
                 surv = FALSE,
                 time = NULL,
                 status = NULL,
                 quant = NULL,
                 geneId,
                 cls,
                 isPaired = FALSE,
                 pairID = NULL,
                 doGroupComparison = FALSE,
                 grpID = NULL,
                 hasDuplic = FALSE,
                 duplicID = NULL,
                 rvm = TRUE,
                 nperm.GSA = 200,
           #      doBiocarta = TRUE,
                 geneSetType = c("BioCarta","KEGG","TF", "microRNA"),
                 fromKEGGdb = TRUE,
                 frommiRTarBasev4 = TRUE,
                 pathwayMin = 5,
                 pathwayMax = 1000,
                 isSingleChannel = T,
                 alpha = 0.05,
                 seed = 123456, # random seed for permutation tests
                 organism = c("human","mouse"),
                 corrtest.method = c("pearson","spearman"),
                 projectPath,
                 outputName = "GeneSetClassComparison",
                 popHTML = TRUE){

  organism <- match.arg(organism)
  corrtest.method <- match.arg(corrtest.method)
  geneSetType <- match.arg(geneSetType)

  if (missing(expr))
    stop("Please provide the expression data.")

  if (missing(filter))
    stop("Please provide the filtering information.")

  if ((surv && is.null(time)) || (surv && is.null(status)))
    stop("Please provide survival time and status.")

  if (organism == "mouse" && !doBiocarta)
    stop("Only BioCarta pathways for mouse are available.")

  if (isPaired && is.null(pairID))
    stop("Please provide the pair IDs.")

  if (doGroupComparison && is.null(grpID))
    stop("Please provide group IDs for group comparison.")

  if (hasDuplic && is.null(duplicID))
    stop("Please provide duplic IDs to average duplicated arrays.")

  if (missing(projectPath))
    projectPath <- file.path(Sys.getenv("R_USER"), "outputFromGSEA")

  if (!surv) time <- status <- NULL

  doqt <- ifelse(!is.null(quant), TRUE, FALSE)

  if (tolower(geneSetType) == "biocarta" || tolower(geneSetType) == "kegg")
    geneSetGroup <- "pathway" else
      if (tolower(geneSetType) == "tf")
        geneSetGroup <- "TranscriptionFactor" else
          if (tolower(geneSetType) == "microrna")
            geneSetGroup <- "microRNA"

  doBiocarta <- ifelse(tolower(geneSetType) == "biocarta", TRUE, FALSE)
  doKEGG <- ifelse(tolower(geneSetType) == "kegg", TRUE, FALSE)
  doTF <- ifelse(tolower(geneSetType) == "tf", TRUE, FALSE)
  doMicroRNA <- ifelse(tolower(geneSetType) == "microrna", TRUE, FALSE)

  if (surv | doqt) {
    cls <- NULL
    rvm <- FALSE
    doGroupComparison <- FALSE
    isPaired <- FALSE
  }

  ArrayToolsPath <- path.package("GSEA")
  WorkPath <- file.path(projectPath, "Fortran", outputName)
  if (!dir.exists(WorkPath)) dir.create(WorkPath, recursive = TRUE)

  successfile <- paste(WorkPath, '/success.txt', sep='')
  fortranerrorfile <- paste(WorkPath, '/fortranerror.txt', sep='')
  unlink(successfile); unlink(fortranerrorfile)



 # outputPath <- file.path(projectPath, "Fortran", outputName)
  # create <class_perm_params.txt>, <LR_data.txt> etc.
 GeneSets <-  prepareGSEA(expr,filter,surv,time,status,doqt,quant, geneId,#geneId,
                          cls,isPaired,pairID,doGroupComparison,grpID,
                          hasDuplic,duplicID,rvm,nperm.GSA,doBiocarta,doKEGG,fromKEGGdb,
                          pathwayMin,pathwayMax,organism,alpha,geneSetGroup,frommiRTarBasev4,WorkPath)


  if (!doqt){
    # create <RunFortran.bat> file
    createfortran.gsea(ArrayToolsPath, WorkPath)
    # run <RunFortrn.bat>
    shell(paste0("\"", WorkPath, "/RunFortran.bat", "\""), wait = TRUE)
  } else{


      expr2 <- expr[filter==1,]
      pCorr <- apply(expr2,1,function (x) cor.test(quant, as.numeric(x), method = corrtest.method)$p.value)
      nvar <- nrow(expr2)
      NumGenesInList <- unlist(lapply(GeneSets,length))
      geneSet <- unlist(GeneSets)
      numSets <- length(NumGenesInList)
      cat('... Performing gene set enrichment analysis ...\n', file=stderr())
      set.seed(seed)
      outc <- .C(geneSetEnrichmentQT,
                 as.integer(nvar),
                 as.integer(geneSet-1),
                 as.integer(NumGenesInList),
                 as.integer(numSets),
              #   as.integer(seed),
                 as.double(pCorr),
                 pvalueSetLS=double(numSets),
                 pvalueSetKS=double(numSets),
                 LS = double(numSets),
                 KS = double(numSets))

     DEG <- data.frame(Index = 1:numSets,
                       nGenesInGeneSet = NumGenesInList,
                       statLS = outc$LS,
                       pLS = outc$pvalueSetLS,
                       statKS = outc$KS,
                       pKS = outc$pvalueSetKS)

      file.DEG <- paste0(WorkPath,"/DEGs.txt")
      unlink(file.DEG)
      cat("Max Number of significant GeneSet categories:\t", file = file.DEG, append = TRUE)
      cat(numSets, file = file.DEG, append = TRUE)

      cat("\n", file = file.DEG, append = TRUE)
      cat("Index\tnGenesInGeneSet\tstatLS\tpLS\tstatKS\tpKS\n", file = file.DEG, append = TRUE)
      for (i in 1:numSets)
        cat(paste0(paste(DEG[i,],collapse="\t"),"\n"), file = file.DEG, append = TRUE)

      file.Pvalues <- paste0(WorkPath,"/Pvalues.txt")
      unlink(file.Pvalues)
      mpv <- cbind(1:nrow(expr2),pCorr)
      colnames(mpv) <- c("index","p")
      write.table(mpv, file = file.Pvalues, sep = "\t", row.names = FALSE, col.names = TRUE, quote = FALSE)
  }

  nGenesForRVM <- nrow(expr)
  nFilteredGenes <-sum(filter)
  if (doBiocarta)
    pathwayType <- "Biocarta" else
      if (doKEGG)
        pathwayType <- "Kegg" else
          if (doTF)
            pathwayType <- "TF" else
              if (doMicroRNA)
                pathwayType <- "microRNA"
  geneSetTable <- GetGeneSetComparisonResults(cls,
                              surv,
                              isPaired,
                              pairID,
                              geneId,
                              filter,
                              nFilteredGenes,
                              nGenesForRVM,
                              doGroupComparison,
                              grpID,
                              doqt,
                              corrtest.method,
                              rvm,
                              nperm.GSA,
                              alpha,
                              isSingleChannel,
                              pathwayType,
                              fromKEGGdb,
                              frommiRTarBasev4,
                              geneSetGroup,
                              organism,
                              projectPath,
                              outputName)

  if (popHTML)
    browseURL(file.path(projectPath, "Output", outputName,
                        paste0(outputName, ".html")))

  cat(paste0("An html document with significantly enriched pathways was saved in",
             file.path(projectPath, "Output", outputName,paste0(outputName, ".html","\n"))))

  return(geneSetTable)
}


createfortran.gsea <- function(ArrayToolsPath, outputpath) {
  # Purpose: create <RunFortran.bat>

  # replace one forward slash with one backward slash for running in COMMAND PROMPT
  outputpath2 <- gsub("/", "\\\\", outputpath)
  ArrayToolsPath <- normalizePath(ArrayToolsPath)
  out <- c(substring(ArrayToolsPath, 1, 2),
           paste0("cd \"", ArrayToolsPath, "\\Fortran\""),
           paste0("GeneSetComparison.exe^\n \"",
                  outputpath2, "\"^\n 1> \"",
                  outputpath2, "\\FortranOutput.txt\"^\n 2> \"",
                  outputpath2, "\\FortranError.txt\""))
  cat(out, file = file.path(outputpath, "RunFortran.bat"), sep = "\n")
}

